@@ -1,15 +1,13 @@
 /* pngquant.c - quantize the colors in an alphamap down to a specified number
 **
-** © 2009-2018 by Kornel Lesiński.
+** © 2009-2019 by Kornel Lesiński.
 ** © 1989, 1991 by Jef Poskanzer.
 ** © 1997-2002 by Greg Roelofs; based on an idea by Stefan Schneider.
 **
 ** See COPYRIGHT file for license.
 */
 
-#define PNGQUANT_VERSION LIQ_VERSION_STRING " (January 2018)"
-
-#define PNGQUANT_USAGE "\
+char *PNGQUANT_USAGE = "\
 usage:  pngquant [options] [ncolors] -- pngfile [pngfile ...]\n\
         pngquant [options] [ncolors] - >stdout <stdin\n\n\
 options:\n\
@@ -18,7 +16,7 @@ options:\n\
   --output file     destination file path to use instead of --ext (synonym: -o)\n\
   --ext new.png     set custom suffix/extension for output filenames\n\
   --quality min-max don't save below min, use fewer colors below max (0-100)\n\
-  --speed N         speed/quality trade-off. 1=slow, 3=default, 11=fast & rough\n\
+  --speed N         speed/quality trade-off. 1=slow, 4=default, 11=fast & rough\n\
   --nofs            disable Floyd-Steinberg dithering\n\
   --posterize N     output lower-precision color (e.g. for ARGB4444 output)\n\
   --strip           remove optional metadata (default on Mac)\n\
@@ -31,7 +29,7 @@ input is stdin, in which case the quantized image will go to stdout).\n\
 If you pass the special output path \"-\" and a single input file, that file\n\
 will be processed and the quantized image will go to stdout.\n\
 The default behavior if the output file exists is to skip the conversion;\n\
-use --force to overwrite. See man page for full list of options.\n"
+use --force to overwrite. See man page for full list of options.\n";
 
 
 #include <stdio.h>
@@ -42,10 +40,11 @@ use --force to overwrite. See man page for full list of options.\n"
 #include <math.h>
 
 #if defined(_WIN32) || defined(WIN32) || defined(__WIN32__)
-#  include <fcntl.h>    /* O_BINARY */
-#  include <io.h>   /* setmode() */
+#include <fcntl.h>    /* O_BINARY */
+#include <io.h>   /* setmode() */
+#include <locale.h> /* UTF-8 locale */
 #else
-#  include <unistd.h>
+#include <unistd.h>
 #endif
 
 #ifdef _OPENMP
@@ -59,14 +58,16 @@ use --force to overwrite. See man page for full list of options.\n"
 #include "libimagequant.h" /* if it fails here, run: git submodule update; ./configure; or add -Ilib to compiler flags */
 #include "pngquant_opts.h"
 
+char *PNGQUANT_VERSION = LIQ_VERSION_STRING " (September 2021)";
+
 static pngquant_error prepare_output_image(liq_result *result, liq_image *input_image, rwpng_color_transform tag, png8_image *output_image);
 static void set_palette(liq_result *result, png8_image *output_image);
 static pngquant_error read_image(liq_attr *options, const char *filename, int using_stdin, png24_image *input_image_p, liq_image **liq_image_p, bool keep_input_pixels, bool strip, bool verbose);
-static pngquant_error write_image(png8_image *output_image, png24_image *output_image24, const char *outname, struct pngquant_options *options);
+static pngquant_error write_image(png8_image *output_image, png24_image *output_image24, const char *outname, struct pngquant_options *options, liq_attr *liq);
 static char *add_filename_extension(const char *filename, const char *newext);
 static bool file_exists(const char *outname);
 
-static void verbose_printf(struct pngquant_options *context, const char *fmt, ...)
+static void verbose_printf(liq_attr *liq, struct pngquant_options *context, const char *fmt, ...)
 {
     if (context->log_callback) {
         va_list va;
@@ -83,7 +84,7 @@ static void verbose_printf(struct pngquant_options *context, const char *fmt, ..
         vsnprintf(buf, required_space, fmt, va);
         va_end(va);
 
-        context->log_callback(context->liq, buf, context->log_callback_user_info);
+        context->log_callback(liq, buf, context->log_callback_user_info);
 #if defined(_MSC_VER)
         free(buf);
 #endif
@@ -126,9 +127,8 @@ static void log_callback_buferred(const liq_attr *attr, const char *msg, void* c
 }
 #endif
 
-static void print_full_version(FILE *fd)
-{
-    fprintf(fd, "pngquant, %s, by Kornel Lesinski, Greg Roelofs.\n"
+void pngquant_internal_print_config(FILE *fd) {
+    fputs(""
         #ifndef NDEBUG
                     "   WARNING: this is a DEBUG (slow) version.\n" /* NDEBUG disables assert() */
         #endif
@@ -138,7 +138,21 @@ static void print_full_version(FILE *fd)
         #if _OPENMP
                     "   Compiled with OpenMP (multicore support).\n"
         #endif
-        , PNGQUANT_VERSION);
+    , fd);
+    fflush(fd);
+}
+
+FILE *pngquant_c_stderr() {
+    return stderr;
+}
+FILE *pngquant_c_stdout() {
+    return stdout;
+}
+
+static void print_full_version(FILE *fd)
+{
+    fprintf(fd, "pngquant, %s, by Kornel Lesinski, Greg Roelofs.\n", PNGQUANT_VERSION);
+    pngquant_internal_print_config(fd);
     rwpng_version_info(fd);
     fputs("\n", fd);
 }
@@ -185,8 +199,8 @@ static bool parse_quality(const char *quality, liq_attr *options, bool *min_qual
     return LIQ_OK == liq_set_quality(options, limit, target);
 }
 
-pngquant_error pngquant_main(struct pngquant_options *options);
-static pngquant_error pngquant_file_internal(const char *filename, const char *outname, struct pngquant_options *options);
+pngquant_error pngquant_main_internal(struct pngquant_options *options, liq_attr *liq);
+static pngquant_error pngquant_file_internal(const char *filename, const char *outname, struct pngquant_options *options, liq_attr *liq);
 
 #ifndef PNGQUANT_NO_MAIN
 int main(int argc, char *argv[])
@@ -201,105 +215,126 @@ int main(int argc, char *argv[])
         return retval;
     }
 
-    return pngquant_main(&options);
-}
-#endif
-
-pngquant_error pngquant_main(struct pngquant_options *options)
-{
-    if (options->print_version) {
+    if (options.print_version) {
         puts(PNGQUANT_VERSION);
         return SUCCESS;
     }
 
-    if (options->missing_arguments) {
+    if (options.missing_arguments) {
         print_full_version(stderr);
         print_usage(stderr);
         return MISSING_ARGUMENT;
     }
 
-    if (options->print_help) {
+    if (options.print_help) {
         print_full_version(stdout);
         print_usage(stdout);
         return SUCCESS;
     }
 
-    options->liq = liq_attr_create();
+#if defined(_WIN32) || defined(WIN32) || defined(__WIN32__)
+    setlocale(LC_ALL, ".65001"); // issue #376; set UTF-8 for Unicode filenames
+#endif
 
-    if (!options->liq) {
+    liq_attr *liq = liq_attr_create();
+
+    if (!liq) {
         fputs("SSE-capable CPU is required for this build.\n", stderr);
         return WRONG_ARCHITECTURE;
     }
 
-    if (options->verbose) {
-        liq_set_log_callback(options->liq, log_callback, NULL);
-        options->log_callback = log_callback;
-    }
-
-    if (options->quality && !parse_quality(options->quality, options->liq, &options->min_quality_limit)) {
+    if (options.quality && !parse_quality(options.quality, liq, &options.min_quality_limit)) {
         fputs("Quality should be in format min-max where min and max are numbers in range 0-100.\n", stderr);
         return INVALID_ARGUMENT;
     }
 
-    if (options->iebug) {
+    if (options.iebug) {
         // opacities above 238 will be rounded up to 255, because IE6 truncates <255 to 0.
-        liq_set_min_opacity(options->liq, 238);
+        liq_set_min_opacity(liq, 238);
         fputs("  warning: the workaround for IE6 is deprecated\n", stderr);
     }
 
-    if (options->last_index_transparent) {
-        liq_set_last_index_transparent(options->liq, true);
+    if (options.verbose) {
+        liq_set_log_callback(liq, log_callback, NULL);
+        options.log_callback = log_callback;
     }
 
-    if (options->speed >= 10) {
-        options->fast_compression = true;
-        if (options->speed == 11) {
-            options->floyd = 0;
-            options->speed = 10;
+    if (options.last_index_transparent) {
+        liq_set_last_index_transparent(liq, true);
+    }
+
+    if (options.speed >= 10) {
+        options.fast_compression = true;
+        if (options.speed == 11) {
+            options.floyd = 0;
+            options.speed = 10;
         }
     }
 
-    if (options->speed && LIQ_OK != liq_set_speed(options->liq, options->speed)) {
+    if (options.speed && LIQ_OK != liq_set_speed(liq, options.speed)) {
         fputs("Speed should be between 1 (slow) and 11 (fast).\n", stderr);
         return INVALID_ARGUMENT;
     }
-    if (options->colors && LIQ_OK != liq_set_max_colors(options->liq, options->colors)) {
+
+    if (options.colors && LIQ_OK != liq_set_max_colors(liq, options.colors)) {
         fputs("Number of colors must be between 2 and 256.\n", stderr);
         return INVALID_ARGUMENT;
     }
 
-    if (options->posterize && LIQ_OK != liq_set_min_posterization(options->liq, options->posterize)) {
+    if (options.posterize && LIQ_OK != liq_set_min_posterization(liq, options.posterize)) {
         fputs("Posterization should be number of bits in range 0-4.\n", stderr);
         return INVALID_ARGUMENT;
     }
 
-    if (options->extension && options->output_file_path) {
+    if (options.extension && options.output_file_path) {
         fputs("--ext and --output options can't be used at the same time\n", stderr);
         return INVALID_ARGUMENT;
     }
 
     // new filename extension depends on options used. Typically basename-fs8.png
-    if (options->extension == NULL) {
-        options->extension = options->floyd > 0 ? "-fs8.png" : "-or8.png";
+    if (options.extension == NULL) {
+        options.extension = options.floyd > 0 ? "-fs8.png" : "-or8.png";
     }
 
-    if (options->output_file_path && options->num_files != 1) {
+    if (options.output_file_path && options.num_files != 1) {
         fputs("  error: Only one input file is allowed when --output is used. This error also happens when filenames with spaces are not in quotes.\n", stderr);
         return INVALID_ARGUMENT;
     }
 
-    if (options->using_stdout && !options->using_stdin && options->num_files != 1) {
+    if (options.using_stdout && !options.using_stdin && options.num_files != 1) {
         fputs("  error: Only one input file is allowed when using the special output path \"-\" to write to stdout. This error also happens when filenames with spaces are not in quotes.\n", stderr);
         return INVALID_ARGUMENT;
     }
 
+    if (!options.num_files && !options.using_stdin) {
+        fputs("No input files specified.\n", stderr);
+        if (options.verbose) {
+            print_full_version(stderr);
+        }
+        print_usage(stderr);
+        return MISSING_ARGUMENT;
+    }
+
+    retval = pngquant_main_internal(&options, liq);
+    liq_attr_destroy(liq);
+    return retval;
+}
+#endif
+
+// Don't use this. This is not a public API.
+pngquant_error pngquant_main_internal(struct pngquant_options *options, liq_attr *liq)
+{
+#if defined(_WIN32) || defined(WIN32) || defined(__WIN32__)
+    setlocale(LC_ALL, ".65001"); // issue #376; set UTF-8 for Unicode filenames
+#endif
+
     if (options->map_file) {
         png24_image tmp = {.width=0};
-        if (SUCCESS != read_image(options->liq, options->map_file, false, &tmp, &options->fixed_palette_image, true, true, false)) {
+        if (SUCCESS != read_image(liq, options->map_file, false, &tmp, &options->fixed_palette_image, true, true, false)) {
             fprintf(stderr, "  error: unable to load %s", options->map_file);
             return INVALID_ARGUMENT;
         }
-        liq_result *tmp_quantize = liq_quantize_image(options->liq, options->fixed_palette_image);
+        liq_result *tmp_quantize = liq_quantize_image(liq, options->fixed_palette_image);
         const liq_palette *pal = liq_get_palette(tmp_quantize);
         if (!pal) {
             fprintf(stderr, "  error: unable to read colors from %s", options->map_file);
@@ -311,15 +346,6 @@ pngquant_error pngquant_main(struct pngquant_options *options)
         liq_result_destroy(tmp_quantize);
     }
 
-    if (!options->num_files && !options->using_stdin) {
-        fputs("No input files specified.\n", stderr);
-        if (options->verbose) {
-            print_full_version(stderr);
-        }
-        print_usage(stderr);
-        return MISSING_ARGUMENT;
-    }
-
 #ifdef _OPENMP
     // if there's a lot of files, coarse parallelism can be used
     if (options->num_files > 2*omp_get_max_threads()) {
@@ -338,14 +364,14 @@ pngquant_error pngquant_main(struct pngquant_options *options)
     for(int i=0; i < options->num_files; i++) {
         const char *filename = options->using_stdin ? "stdin" : options->files[i];
         struct pngquant_options opts = *options;
-        opts.liq = liq_attr_copy(options->liq);
+        liq_attr *local_liq = liq_attr_copy(liq);
 
 
         #ifdef _OPENMP
         struct buffered_log buf = {0};
         if (opts.log_callback && omp_get_num_threads() > 1 && opts.num_files > 1) {
-            liq_set_log_callback(opts.liq, log_callback_buferred, &buf);
-            liq_set_log_flush_callback(opts.liq, log_callback_buferred_flush, &buf);
+            liq_set_log_callback(local_liq, log_callback_buferred, &buf);
+            liq_set_log_flush_callback(local_liq, log_callback_buferred_flush, &buf);
             opts.log_callback = log_callback_buferred;
             opts.log_callback_user_info = &buf;
         }
@@ -367,12 +393,12 @@ pngquant_error pngquant_main(struct pngquant_options *options)
         }
 
         if (SUCCESS == retval) {
-            retval = pngquant_file_internal(filename, outname, &opts);
+            retval = pngquant_file_internal(filename, outname, &opts, local_liq);
         }
 
         free(outname_free);
 
-        liq_attr_destroy(opts.liq);
+        liq_attr_destroy(local_liq);
 
         if (retval) {
             #pragma omp critical
@@ -389,61 +415,60 @@ pngquant_error pngquant_main(struct pngquant_options *options)
     }
 
     if (error_count) {
-        verbose_printf(options, "There were errors quantizing %d file%s out of a total of %d file%s.",
+        verbose_printf(liq, options, "There were errors quantizing %d file%s out of a total of %d file%s.",
                        error_count, (error_count == 1)? "" : "s", file_count, (file_count == 1)? "" : "s");
     }
     if (skipped_count) {
-        verbose_printf(options, "Skipped %d file%s out of a total of %d file%s.",
+        verbose_printf(liq, options, "Skipped %d file%s out of a total of %d file%s.",
                        skipped_count, (skipped_count == 1)? "" : "s", file_count, (file_count == 1)? "" : "s");
     }
     if (!skipped_count && !error_count) {
-        verbose_printf(options, "Quantized %d image%s.",
+        verbose_printf(liq, options, "Quantized %d image%s.",
                        file_count, (file_count == 1)? "" : "s");
     }
 
     if (options->fixed_palette_image) liq_image_destroy(options->fixed_palette_image);
-    liq_attr_destroy(options->liq);
 
     return latest_error;
 }
 
 /// Don't hack this. Instead use https://github.com/ImageOptim/libimagequant/blob/f54d2f1a3e1cf728e17326f4db0d45811c63f063/example.c
-static pngquant_error pngquant_file_internal(const char *filename, const char *outname, struct pngquant_options *options)
+static pngquant_error pngquant_file_internal(const char *filename, const char *outname, struct pngquant_options *options, liq_attr *liq)
 {
     pngquant_error retval = SUCCESS;
 
-    verbose_printf(options, "%s:", filename);
+    verbose_printf(liq, options, "%s:", filename);
 
     liq_image *input_image = NULL;
     png24_image input_image_rwpng = {.width=0};
     bool keep_input_pixels = options->skip_if_larger || (options->using_stdout && options->min_quality_limit); // original may need to be output to stdout
     if (SUCCESS == retval) {
-        retval = read_image(options->liq, filename, options->using_stdin, &input_image_rwpng, &input_image, keep_input_pixels, options->strip, options->verbose);
+        retval = read_image(liq, filename, options->using_stdin, &input_image_rwpng, &input_image, keep_input_pixels, options->strip, options->verbose);
     }
 
     int quality_percent = 90; // quality on 0-100 scale, updated upon successful remap
     png8_image output_image = {.width=0};
     if (SUCCESS == retval) {
-        verbose_printf(options, "  read %luKB file", (input_image_rwpng.file_size+1023UL)/1024UL);
+        verbose_printf(liq, options, "  read %luKB file", (input_image_rwpng.file_size+1023UL)/1024UL);
 
         if (RWPNG_ICCP == input_image_rwpng.input_color) {
-            verbose_printf(options, "  used embedded ICC profile to transform image to sRGB colorspace");
+            verbose_printf(liq, options, "  used embedded ICC profile to transform image to sRGB colorspace");
         } else if (RWPNG_GAMA_CHRM == input_image_rwpng.input_color) {
-            verbose_printf(options, "  used gAMA and cHRM chunks to transform image to sRGB colorspace");
+            verbose_printf(liq, options, "  used gAMA and cHRM chunks to transform image to sRGB colorspace");
         } else if (RWPNG_ICCP_WARN_GRAY == input_image_rwpng.input_color) {
-            verbose_printf(options, "  warning: ignored ICC profile in GRAY colorspace");
+            verbose_printf(liq, options, "  warning: ignored ICC profile in GRAY colorspace");
         } else if (RWPNG_COCOA == input_image_rwpng.input_color) {
             // No comment
         } else if (RWPNG_SRGB == input_image_rwpng.input_color) {
-            verbose_printf(options, "  passing sRGB tag from the input");
+            verbose_printf(liq, options, "  passing sRGB tag from the input");
         } else if (input_image_rwpng.gamma != 0.45455) {
-            verbose_printf(options, "  converted image from gamma %2.1f to gamma 2.2",
+            verbose_printf(liq, options, "  converted image from gamma %2.1f to gamma 2.2",
                            1.0/input_image_rwpng.gamma);
         }
 
         // when using image as source of a fixed palette the palette is extracted using regular quantization
         liq_result *remap;
-        liq_error remap_error = liq_image_quantize(options->fixed_palette_image ? options->fixed_palette_image : input_image, options->liq, &remap);
+        liq_error remap_error = liq_image_quantize(options->fixed_palette_image ? options->fixed_palette_image : input_image, liq, &remap);
 
         if (LIQ_OK == remap_error) {
 
@@ -463,7 +488,7 @@ static pngquant_error pngquant_file_internal(const char *filename, const char *o
                 double palette_error = liq_get_quantization_error(remap);
                 if (palette_error >= 0) {
                     quality_percent = liq_get_quantization_quality(remap);
-                    verbose_printf(options, "  mapped image to new colors...MSE=%.3f (Q=%d)", palette_error, quality_percent);
+                    verbose_printf(liq, options, "  mapped image to new colors...MSE=%.3f (Q=%d)", palette_error, quality_percent);
                 }
             }
             liq_result_destroy(remap);
@@ -479,7 +504,7 @@ static pngquant_error pngquant_file_internal(const char *filename, const char *o
         if (options->skip_if_larger) {
             // this is very rough approximation, but generally avoid losing more quality than is gained in file size.
             // Quality is raised to 1.5, because even greater savings are needed to justify big quality loss.
-            // but >50% savings are considered always worthwile in order to allow low quality conversions to work at all
+            // but >50% savings are considered always worthwhile in order to allow low quality conversions to work at all
             const double quality = quality_percent/100.0;
             const double expected_reduced_size = pow(quality, 1.5);
             output_image.maximum_file_size = (input_image_rwpng.file_size-1) * (expected_reduced_size < 0.5 ? 0.5 : expected_reduced_size);
@@ -487,20 +512,20 @@ static pngquant_error pngquant_file_internal(const char *filename, const char *o
 
         output_image.fast_compression = options->fast_compression;
         output_image.chunks = input_image_rwpng.chunks; input_image_rwpng.chunks = NULL;
-        retval = write_image(&output_image, NULL, outname, options);
+        retval = write_image(&output_image, NULL, outname, options, liq);
 
         if (TOO_LARGE_FILE == retval) {
-            verbose_printf(options, "  file exceeded expected size of %luKB", (unsigned long)output_image.maximum_file_size/1024UL);
+            verbose_printf(liq, options, "  file exceeded expected size of %luKB", (unsigned long)output_image.maximum_file_size/1024UL);
         }
         if (SUCCESS == retval && output_image.metadata_size > 0) {
-            verbose_printf(options, "  copied %dKB of additional PNG metadata", (int)(output_image.metadata_size+999)/1000);
+            verbose_printf(liq, options, "  copied %dKB of additional PNG metadata", (int)(output_image.metadata_size+999)/1000);
         }
     }
 
     if (options->using_stdout && keep_input_pixels && (TOO_LARGE_FILE == retval || TOO_LOW_QUALITY == retval)) {
         // when outputting to stdout it'd be nasty to create 0-byte file
         // so if quality is too low, output 24-bit original
-        pngquant_error write_retval = write_image(NULL, &input_image_rwpng, outname, options);
+        pngquant_error write_retval = write_image(NULL, &input_image_rwpng, outname, options, liq);
         if (write_retval) {
             retval = write_retval;
         }
@@ -545,8 +570,8 @@ static char *add_filename_extension(const char *filename, const char *newext)
     char* outname = malloc(x+4+strlen(newext)+1);
     if (!outname) return NULL;
 
-    strncpy(outname, filename, x);
-    if (strncmp(outname+x-4, ".png", 4) == 0 || strncmp(outname+x-4, ".PNG", 4) == 0) {
+    strcpy(outname, filename);
+    if (x > 4 && (strncmp(outname+x-4, ".png", 4) == 0 || strncmp(outname+x-4, ".PNG", 4) == 0)) {
         strcpy(outname+x-4, newext);
     } else {
         strcpy(outname+x, newext);
@@ -594,7 +619,7 @@ static bool replace_file(const char *from, const char *to, const bool force) {
     return (0 == rename(from, to));
 }
 
-static pngquant_error write_image(png8_image *output_image, png24_image *output_image24, const char *outname, struct pngquant_options *options)
+static pngquant_error write_image(png8_image *output_image, png24_image *output_image24, const char *outname, struct pngquant_options *options, liq_attr *liq)
 {
     FILE *outfile;
     char *tempname = NULL;
@@ -604,9 +629,9 @@ static pngquant_error write_image(png8_image *output_image, png24_image *output_
         outfile = stdout;
 
         if (output_image) {
-            verbose_printf(options, "  writing %d-color image to stdout", output_image->num_palette);
+            verbose_printf(liq, options, "  writing %d-color image to stdout", output_image->num_palette);
         } else {
-            verbose_printf(options, "  writing truecolor image to stdout");
+            verbose_printf(liq, options, "  writing truecolor image to stdout");
         }
     } else {
         tempname = temp_filename(outname);
@@ -619,9 +644,9 @@ static pngquant_error write_image(png8_image *output_image, png24_image *output_
         }
 
         if (output_image) {
-            verbose_printf(options, "  writing %d-color image as %s", output_image->num_palette, filename_part(outname));
+            verbose_printf(liq, options, "  writing %d-color image as %s", output_image->num_palette, filename_part(outname));
         } else {
-            verbose_printf(options, "  writing truecolor image as %s", filename_part(outname));
+            verbose_printf(liq, options, "  writing truecolor image as %s", filename_part(outname));
         }
     }
 
@@ -714,8 +739,8 @@ static pngquant_error prepare_output_image(liq_result *result, liq_image *input_
     ** Step 3.7 [GRR]: allocate memory for the entire indexed image
     */
 
-    output_image->indexed_data = malloc(output_image->height * output_image->width);
-    output_image->row_pointers = malloc(output_image->height * sizeof(output_image->row_pointers[0]));
+    output_image->indexed_data = malloc((size_t)output_image->height * (size_t)output_image->width);
+    output_image->row_pointers = malloc((size_t)output_image->height * sizeof(output_image->row_pointers[0]));
 
     if (!output_image->indexed_data || !output_image->row_pointers) {
         return OUT_OF_MEMORY_ERROR;
